use crate::hash::hash_u32;

/// Performs a base-2 Owen scramble on an integer.
pub fn owen(n: u32, seed: u32) -> u32 {
    let mut result = n;

    for i in 0..32 {
        let mask = if i < 31 { (!0) << (i + 1) } else { 0 };
        result ^= hash_u32(n & mask, seed) & (1 << i);
    }

    result
}

/// Performs z-scrambling from the paper "Screen-Space Blue-Noise
/// Diffusion of Monte Carlo Sampling Error via Hierarchical Ordering
/// of Pixels" by Ahmed et al., except using a more complete and
/// efficient algorithm based on the Owen scrambling function above.
pub fn z_scramble(n: u32, seed: u32) -> u32 {
    let mut result = 0;

    for i in 0..16 {
        let mask = if i < 15 { (!0) << ((i + 1) * 2) } else { 0 };
        let entry = (hash_u32(n & mask, seed) % 24) as usize;
        let cell = ((n >> (i * 2)) & 0b11) as usize;
        result |= (Z_TABLE[entry][cell] as u32) << (i * 2);
    }

    result
}

const Z_TABLE: &[[u8; 4]; 24] = &[
    [0, 1, 2, 3],
    [0, 1, 3, 2],
    [0, 2, 1, 3],
    [0, 2, 3, 1],
    [0, 3, 1, 2],
    [0, 3, 2, 1],
    [1, 0, 2, 3],
    [1, 0, 3, 2],
    [1, 2, 0, 3],
    [1, 2, 3, 0],
    [1, 3, 0, 2],
    [1, 3, 2, 0],
    [2, 0, 1, 3],
    [2, 0, 3, 1],
    [2, 1, 0, 3],
    [2, 1, 3, 0],
    [2, 3, 0, 1],
    [2, 3, 1, 0],
    [3, 0, 1, 2],
    [3, 0, 3, 2],
    [3, 1, 0, 2],
    [3, 1, 2, 0],
    [3, 2, 0, 1],
    [3, 2, 1, 0],
];
